#version 330

precision highp float; // Video card drivers require this line to function properly

in vec4 viewspaceNormal;
in vec4 viewspacePosition;
in vec4 eyeVector;
in vec3 minusLightDir;
in vec4 lightSpacePosition;
in vec2 texCoord;

uniform sampler2D activeTexture;
uniform sampler2D shadowMap;

uniform int shadingCheck;

out vec4 fragColor;

// Variance Shadow Mapping globals
vec4 ShadowCoordPostW;
const float lightBleedingReductionAmount = 0.17; //0.0 means no bleeding reduction
const float minVariance = 0.00000005;
const float bias = 0.0002;
float shadowMapSampleSize = 1.0/2048.0;

float linstep(float min, float max, float v)
{
	return clamp( (v - min) / (max - min) , 0.0, 1.0 );
}

float reduceLightBleeding(float p_max, float amount)
{
	// Remove the [0, Amount] tail and linearly rescale (Amount, 1].
	return linstep(amount, 1.0, p_max);
}

/// offset lookup function
vec2 offset_lookup(sampler2D map, vec4 loc, vec2 offset){
	return textureProj(map, vec4(vec2(loc.xy + offset*shadowMapSampleSize*loc.w), loc.z, loc.w), 0.0).rg;
}

float chebyshevUpperBound(float lightSpaceDepth)
{

	vec4 textureSpaceCoord = ShadowCoordPostW*0.5 + 0.5;
	vec2 shadowMapSample;

	// use 1 lookup
	shadowMapSample    = textureProj(shadowMap, textureSpaceCoord, 0.0).rg;

	float firstMoment		= shadowMapSample.r;
	float secondMoment		= shadowMapSample.g;
	
	float mean			= firstMoment;
	float variance		= secondMoment - firstMoment*firstMoment;
	
	/**
	* "Note that it is useful to clamp the minimum variance to a very small value,
	* as we have done here, to avoid any numeric issues that may occur during filtering.
	* Additionally, simply clamping the variance often eliminates shadow biasing issues"
	**/
	variance = max(variance, minVariance);
	
	/**
	* "One-tailed inequality only valid if distance > mean"
	**/
	if (lightSpaceDepth < mean){
		return 1.0;
	}
	
	/** 
	* "Using the variance, we can apply Chebyshev's inequality
	* to compute an upper bound on the probability that
	* the currently shaded surface (at depth distance) is occluded"
	**/
	float p_max = (variance) /
	( (variance) + (lightSpaceDepth - mean)*(lightSpaceDepth - mean) );
	
	return reduceLightBleeding(p_max, lightBleedingReductionAmount);
}

void main(void)
{
	// defines
	vec4 objColor = texture(activeTexture, texCoord);
	vec4 specColorMaterial = vec4(1.0, 1.0, 1.0, 1.0);
	vec4 specColorLight = vec4(1.0, 1.0, 1.0, 1.0);
	
	if(shadingCheck == 10) fragColor =  objColor;
	else
	{
		// ambient
		vec4 ambient = objColor * 0.18;
	
		// diffuse
		float lambertFactor = max( dot(normalize(viewspaceNormal.xyz), normalize(minusLightDir)), 0.0 );
		vec4 lambert = objColor * lambertFactor;
	
		// specular
		int specularExponent = 20;
		vec3 halfVector = normalize( normalize(eyeVector.xyz) +  normalize(minusLightDir) );
		vec4 Is = specColorMaterial * specColorLight * pow( max( dot( normalize(viewspaceNormal.xyz), halfVector ), 0.0 ), specularExponent );

		// shadow
		ShadowCoordPostW = lightSpacePosition / lightSpacePosition.w;
		float shadowCoeff = chebyshevUpperBound(ShadowCoordPostW.z*0.5 + 0.5 - bias);

		// compositing
		fragColor =  vec4( ambient.rgb + (vec3(lambert) + lambertFactor*vec3(Is)) * shadowCoeff, 1.0 );
	}
}
